@@ -137,6 +137,9 @@ class DatabaseManager:
         if self.connection_manager.config.database == clean_name:
             self.connection_manager.config.database = None
 
+
+
+
     def generate_schema(
         self,
         schema_name: str,
@@ -193,7 +196,7 @@ class DatabaseManager:
             for table, col, typ in filtered_columns:
                 samples: List[Any] = []
 
-                # attempt primary-key range sampling for performance
+                # stratified PK-range sampling: 3/4/3 from three equal buckets
                 try:
                     cursor.execute(
                         "SELECT COLUMN_NAME, DATA_TYPE FROM INFORMATION_SCHEMA.COLUMNS "
@@ -201,34 +204,40 @@ class DatabaseManager:
                         (self.connection_manager.config.database, table)
                     )
                     pk_info = cursor.fetchone()
-                    if pk_info:
-                        pk_col, pk_type = pk_info
-                        if pk_type.upper() in ("INT", "BIGINT", "SMALLINT", "MEDIUMINT", "TINYINT"):
-                            cursor.execute(f"SELECT MIN(`{pk_col}`), MAX(`{pk_col}`) FROM `{table}`")
-                            min_id, max_id = cursor.fetchone()
-                            import random
-                            trials = 0
-                            while len(samples) < num_examples and trials < num_examples * 10:
-                                rand_id = random.randint(min_id, max_id)
-                                cursor.execute(
-                                    f"SELECT `{col}` FROM `{table}` "
-                                    f"WHERE `{pk_col}` >= %s AND `{col}` IS NOT NULL LIMIT 1",
-                                    (rand_id,)
-                                )
-                                row = cursor.fetchone()
-                                if row:
-                                    val = row[0]
-                                    if val not in samples:
-                                        samples.append(val)
-                                trials += 1
-                            logger.debug(
-                                f"Used PK-range sampling for {table}.{col} "
-                                f"with {trials} trials and {len(samples)} samples"
+                    if not pk_info or pk_info[1].upper() not in ("INT","BIGINT","SMALLINT","MEDIUMINT","TINYINT"):
+                        raise ValueError("No suitable integer PK")
+
+                    pk_col = pk_info[0]
+                    cursor.execute(f"SELECT MIN(`{pk_col}`), MAX(`{pk_col}`) FROM `{table}`")
+                    min_id, max_id = cursor.fetchone()
+                    import random
+
+                    total = max_id - min_id + 1
+                    size = total // 3
+                    buckets = [
+                        (min_id,           min_id + size - 1, 3),
+                        (min_id + size,    min_id + 2*size - 1, 4),
+                        (min_id + 2*size,  max_id,             3),
+                    ]
+
+                    for start, end, want in buckets:
+                        got = []
+                        trials = 0
+                        while len(got) < want and trials < want * 10:
+                            trials += 1
+                            rand_id = random.randint(start, end)
+                            cursor.execute(
+                                f"SELECT `{col}` FROM `{table}` "
+                                f"WHERE `{pk_col}` >= %s AND `{col}` IS NOT NULL LIMIT 1",
+                                (rand_id,)
                             )
-                        else:
-                            raise ValueError("Non-integer PK")
-                    else:
-                        raise ValueError("No primary key")
+                            row = cursor.fetchone()
+                            if row and row[0] not in got:
+                                got.append(row[0])
+                        samples.extend(got)
+                        logger.debug(
+                            f"Bucket {start}–{end}: wanted {want}, got {len(got)} after {trials} trials"
+                        )
                 except Exception:
                     # fallback reservoir sampling on first 10000 distinct values
                     cursor.execute(
@@ -247,8 +256,7 @@ class DatabaseManager:
                                 reservoir[j] = val
                     samples = reservoir
                     logger.debug(
-                        f"Used fallback reservoir sampling for {table}.{col} "
-                        f"({len(samples)} samples)"
+                        f"Used fallback reservoir sampling for {table}.{col} ({len(samples)} samples)"
                     )
 
                 # assemble field info
@@ -279,6 +287,7 @@ class DatabaseManager:
 
         return out_path
 
+
     def list_databases(
         self,
         show_fields: bool = False,
@@ -372,4 +381,78 @@ class DatabaseManager:
             return {"data": result, "error": None, "final_sql": final_sql}
         except mysql.connector.Error as e:
             logger.error(f"Error executing query: {e}")
-            return {"data": None, "error": str(e), "final_sql": sql_query}
\ No newline at end of file
+            return {"data": None, "error": str(e), "final_sql": sql_query}
+
+    def ensure_transform_functions(
+        self,
+        initialism_fn: str = "initialism",
+        metaphone_udf: str = "double_metaphone",
+        udf_soname: str = "lib_mysqludf_morph.so"
+    ) -> None:
+        """
+        Ensure that the MySQL functions for initialism (stored function) and metaphone (UDF)
+        exist in the connected database. Creates them if missing.
+        """
+        conn = self.connection_manager.connect()
+        try:
+            with get_cursor(conn) as cursor:
+                # Check for existing stored function
+                cursor.execute(
+                    "SELECT COUNT(*) FROM information_schema.routines "
+                    "WHERE routine_schema=%s AND routine_name=%s AND routine_type='FUNCTION'",
+                    (self.connection_manager.config.database, initialism_fn)
+                )
+                if cursor.fetchone()[0] == 0:
+                    # Drop any existing stub
+                    cursor.execute(f"DROP FUNCTION IF EXISTS `{initialism_fn}`")
+                    conn.commit()
+                    # Use a pure-Python connection to execute multi-statement DDL
+                    cfg = self.connection_manager.config
+                    pure_conn = mysql.connector.connect(
+                        host=cfg.host,
+                        user=cfg.user,
+                        password=cfg.password,
+                        port=cfg.port,
+                        database=cfg.database,
+                        use_pure=True
+                    )
+                    try:
+                        pure_cursor = pure_conn.cursor()
+                        create_fn_sql = f"""
+                        CREATE FUNCTION `{initialism_fn}`(txt TEXT)
+                          RETURNS VARCHAR(255)
+                          DETERMINISTIC
+                        BEGIN
+                          DECLARE result VARCHAR(255) DEFAULT '';
+                          DECLARE word  VARCHAR(255);
+                          DECLARE pos   INT DEFAULT 1;
+                          WHILE pos > 0 DO
+                            SET word = SUBSTRING_INDEX(SUBSTRING(txt, pos), ' ', 1);
+                            IF word = '' THEN
+                              LEAVE;
+                            END IF;
+                            SET result = CONCAT(result, UPPER(LEFT(word, 1)));
+                            SET pos = LOCATE(' ', txt, pos) + 1;
+                            IF pos = 1 THEN
+                              LEAVE;
+                            END IF;
+                          END WHILE;
+                          RETURN result;
+                        END;
+                        """
+                        pure_cursor.execute(create_fn_sql, multi=True)
+                        pure_conn.commit()
+                        logger.info(f"Created stored function `{initialism_fn}`")
+                    finally:
+                        pure_conn.close()
+                # Check for existing UDF
+                cursor.execute(
+                    "SELECT COUNT(*) FROM mysql.func WHERE name = %s",
+                    (metaphone_udf,)
+                )
+                if cursor.fetchone()[0] == 0:
+                    cursor.execute(f"INSTALL SONAME '{udf_soname}'")
+                    conn.commit()
+                    logger.info(f"Installed UDF library for `{metaphone_udf}`")
+        finally:
+            pass
\ No newline at end of file
