@@ -10,6 +10,7 @@ from canonmap.services.db_mysql.schemas import TableField, FieldTransformType
 from canonmap.services.db_mysql.adapters.cursor import get_cursor
 
 from canonmap.services.db_mysql.schemas import CreateDDLResponse # type: ignore
+from canonmap.services.db_mysql.managers.database import DatabaseManager
 
 logger = logging.getLogger(__name__)
 
@@ -161,23 +162,37 @@ class TableManager:
     def create_table_fields(
         self,
         fields: list["TableField"],
+        field_transform: FieldTransformType,
         pk_field: str | None = None,
         batch_size: int = 10_000,
+        skip_existing: bool = True,
     ) -> dict[str, list[str]]:
         """
-        For each TableField(table_name, field_name, field_transform) create a derived column:
+        For each TableField create a derived column with the specified transform:
         initialism  -> FIRST LETTERS ONLY (ABC)
         phonetic    -> Double Metaphone primary code
         soundex     -> MySQL SOUNDEX()
 
+        Args:
+            fields: List of TableField objects to transform
+            field_transform: Type of transform to apply to all fields
+            pk_field: Primary key field name (auto-detected if None)
+            batch_size: Batch size for processing large tables
+            skip_existing: If True, only process NULL/empty cells in transformed fields
+
         Returns: {table_name: [new_field_names]}
         """
-        # Group/validate
+        # ensure the transform functions exist for initialism and phonetic
+        if field_transform in (FieldTransformType.INITIALISM, FieldTransformType.PHONETIC):
+            db_manager = DatabaseManager(self.connection_manager)
+            db_manager.ensure_transform_functions()
+        # Validate transform type
+        if field_transform not in {FieldTransformType.INITIALISM, FieldTransformType.PHONETIC, FieldTransformType.SOUNDEX}:
+            raise ValueError(f"Invalid transform '{field_transform}'")
+        
+        # Group by table
         by_table: dict[str, list[TableField]] = {}
         for f in fields:
-            kind = (f.field_transform or "").lower()
-            if kind not in {"initialism", "phonetic", "soundex"}:
-                raise ValueError(f"Invalid transform '{f.field_transform}' for {f.table_name}.{f.field_name}")
             by_table.setdefault(f.table_name, []).append(f)
 
         conn = self.connection_manager.connect()
@@ -225,14 +240,14 @@ class TableManager:
                 if f.field_name not in cols:
                     raise ValueError(f"Column '{f.field_name}' not found in '{table_name}'")
 
-                new_field = f"__{f.field_name}_{f.field_transform.value.lower()}__"
+                new_field = f"__{f.field_name}_{field_transform.value.lower()}__"
                 if new_field not in cols:
                     with get_cursor(conn) as cursor:
                         cursor.execute(f"ALTER TABLE `{table_name}` ADD COLUMN `{new_field}` VARCHAR(255)")
                     conn.commit()
                     cols.add(new_field)
 
-                if f.field_transform == FieldTransformType.SOUNDEX:
+                if field_transform == FieldTransformType.SOUNDEX:
                     sql_jobs.append((f, new_field))
                 else:
                     py_jobs.append((f, new_field))
@@ -240,16 +255,33 @@ class TableManager:
             # SOUNDEX via SQL
             for f, new_field in sql_jobs:
                 with get_cursor(conn) as cursor:
-                    cursor.execute(
-                        f"UPDATE `{table_name}` SET `{new_field}` = SOUNDEX(`{f.field_name}`)"
-                    )
+                    if skip_existing:
+                        cursor.execute(
+                            f"UPDATE `{table_name}` SET `{new_field}` = SOUNDEX(`{f.field_name}`) "
+                            f"WHERE `{new_field}` IS NULL OR `{new_field}` = ''"
+                        )
+                    else:
+                        cursor.execute(
+                            f"UPDATE `{table_name}` SET `{new_field}` = SOUNDEX(`{f.field_name}`)"
+                        )
                 conn.commit()
 
             # Python transforms batched
             if py_jobs:
-                with get_cursor(conn) as cursor:
-                    cursor.execute(f"SELECT COUNT(*) FROM `{table_name}`")
-                    total_rows = cursor.fetchone()[0]
+                # Count total rows to process
+                if skip_existing:
+                    # Count only rows where transformed fields are NULL/empty
+                    null_conditions = []
+                    for f, new_field in py_jobs:
+                        null_conditions.append(f"`{new_field}` IS NULL OR `{new_field}` = ''")
+                    
+                    with get_cursor(conn) as cursor:
+                        cursor.execute(f"SELECT COUNT(*) FROM `{table_name}` WHERE {' OR '.join(null_conditions)}")
+                        total_rows = cursor.fetchone()[0]
+                else:
+                    with get_cursor(conn) as cursor:
+                        cursor.execute(f"SELECT COUNT(*) FROM `{table_name}`")
+                        total_rows = cursor.fetchone()[0]
 
                 batches = math.ceil(total_rows / batch_size)
                 select_cols = ", ".join({f"`{effective_pk}`"} | {f"`{f.field_name}`" for f, _ in py_jobs})
@@ -257,10 +289,23 @@ class TableManager:
                 for i in range(batches):
                     offset = i * batch_size
                     with get_cursor(conn) as cursor:
-                        cursor.execute(
-                            f"SELECT {select_cols} FROM `{table_name}` LIMIT %s OFFSET %s",
-                            (batch_size, offset)
-                        )
+                        if skip_existing:
+                            # Only select rows where transformed fields are NULL/empty
+                            null_conditions = []
+                            for f, new_field in py_jobs:
+                                null_conditions.append(f"`{new_field}` IS NULL OR `{new_field}` = ''")
+                            
+                            cursor.execute(
+                                f"SELECT {select_cols} FROM `{table_name}` "
+                                f"WHERE {' OR '.join(null_conditions)} "
+                                f"LIMIT %s OFFSET %s",
+                                (batch_size, offset)
+                            )
+                        else:
+                            cursor.execute(
+                                f"SELECT {select_cols} FROM `{table_name}` LIMIT %s OFFSET %s",
+                                (batch_size, offset)
+                            )
                         rows = cursor.fetchall()
                         # capture column order once
                         colnames = [d[0] for d in cursor.description]
@@ -275,7 +320,7 @@ class TableManager:
                         row_map = dict(zip(colnames, row))
                         for f, new_field in py_jobs:
                             src_val = row_map[f.field_name]
-                            if f.field_transform == FieldTransformType.INITIALISM:
+                            if field_transform == FieldTransformType.INITIALISM:
                                 transformed = to_initialism(src_val)
                             else:
                                 transformed = to_phonetic(src_val)
@@ -289,17 +334,20 @@ class TableManager:
                             )
                     conn.commit()
 
-            created[table_name] = [f"{f.field_name}_{f.field_transform.value.lower()}" for f in flist]
+            created[table_name] = [f"{f.field_name}_{field_transform.value.lower()}" for f in flist]
 
         logger.info("✅ Finished create_table_fields.")
         return created
 
-    def drop_table_fields(self, fields: list["TableField"]) -> dict[str, list[str]]:
+    def drop_table_fields(self, fields: list["TableField"], field_transform: FieldTransformType) -> dict[str, list[str]]:
         """
         Drop the derived columns referenced by TableField objects.
-        Uses TableField.field_transform to infer the column name:
+        Uses the provided field_transform to infer the column name:
         new_col = f"{field_name}_{field_transform.lower()}"
-        If field_transform is None, it will try to drop `field_name` directly.
+
+        Args:
+            fields: List of TableField objects
+            field_transform: Type of transform that was applied
 
         Returns: {table_name: [dropped_cols]}
         """
@@ -324,10 +372,7 @@ class TableManager:
             # build list of real cols to drop
             to_drop: list[str] = []
             for f in flist:
-                if f.field_transform:
-                    col = f"{f.field_name}_{f.field_transform.value.lower()}"
-                else:
-                    col = f.field_name
+                col = f"__{f.field_name}_{field_transform.value.lower()}__"
                 if col in cols:
                     to_drop.append(col)
                 else:
